#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# IMPORTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

from src.thirdparty.types import *;
from src.thirdparty.maths import *;

from models.generated.config import *;
from src.core.utils import *;
from src.models.rucksack import *;
from src.models.stacks import *;
from src.algorithms.rucksack.display import *;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# EXPORTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

__all__ = [
    'rucksack_greedy_algorithm',
    'rucksack_branch_and_bound_algorithm',
];

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# METHOD greedy algorithm
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def rucksack_greedy_algorithm(
    max_cost: float,
    costs: np.ndarray,
    values: np.ndarray,
    items: np.ndarray,
    fractional: bool,
    verbose: bool,
) -> Solution:
    '''
    Durch den Greedy-Algorithm wird der optimale Wert eines Rucksacks
    unter Rücksicht der Kapizitätsschranke eingeschätzt.

    NOTE: Wenn man `fractional = True` verwendet, liefert der Algorithmus
    eine obere Schranke des maximalen Wertes beim Originalproblem.
    '''
    # sortiere daten:
    order = get_sort_order(costs=costs, values=values);

    # verbose output hier behandeln (irrelevant für Algorithmus):
    if verbose:
        repr = display_order(order=order, costs=costs, values=values, items=items, one_based=True);
        print('');
        print('\x1b[1mRucksack Problem - Greedy\x1b[0m');
        print('');
        print(repr);
        print('');

    # führe greedy aus:
    n = len(costs);
    cost_total = 0;
    vector = [ 0 for _ in range(n) ];
    for i in order:
        # füge Item i hinzu, solange das Gesamtgewicht noch <= Schranke
        if cost_total + costs[i] <= max_cost:
            cost_total += costs[i];
            vector[i] = 1;
        # falls Bruchteile erlaubt sind, füge einen Bruchteil des i. Items hinzu und abbrechen
        elif fractional:
            vector[i] = (max_cost - cost_total)/costs[i];
            break;
        # ansonsten weiter machen:
        else:
            continue;

    # Aspekte der Lösung speichern:
    rucksack = [i for i, v in enumerate(vector) if v > 0]; # Indexes von Items im Rucksack
    soln = Solution(
        vector = vector,
        items = items[rucksack].tolist(),
        costs = costs[rucksack].tolist(),
        values = values[rucksack].tolist(),
    );

    # verbose output hier behandeln (irrelevant für Algorithmus):
    if verbose:
        repr_rucksack = display_rucksack(items=items[rucksack], costs=costs[rucksack], values=values[rucksack]);
        print('\x1b[1mEingeschätzte Lösung\x1b[0m');
        print('');
        print(f'Mask: {soln.vector}');
        print('Rucksack:')
        print(repr_rucksack);
        print('');

    # Lösung ausgeben
    return soln;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# METHOD branch and bound algorithm
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def rucksack_branch_and_bound_algorithm(
    max_cost: float,
    costs: np.ndarray,
    values: np.ndarray,
    items: np.ndarray,
    verbose: bool,
) -> Solution:
    '''
    Durch Branch & Bound wird der optimale Wert eines Rucksacks
    unter Rücksicht der Kapizitätsschranke exakt und effizienter bestimmt.
    '''

    order = get_sort_order(costs=costs, values=values);

    # verbose output hier behandeln (irrelevant für Algorithmus):
    if verbose:
        repr = display_order(order=order, costs=costs, values=values, items=items, one_based=True);
        print('');
        print('\x1b[1mRucksack Problem - Branch & Bound\x1b[0m');
        print('');
        print(repr);
        print('');

    logged_steps = [];
    vector = empty_mask(n=len(costs));
    lb_estimate = np.inf;
    S = Stack();
    S.push(vector);
    while not S.empty():
        lb, u, can_add_all, can_add_none = estimate_lower_bound(mask=S.top(), max_cost=max_cost, costs=costs, values=values, items=items);
        if verbose:
            logged_steps.append((lb_estimate, lb, str(S), u, can_add_all, can_add_none));
        # Update nur nötig, wenn die (eingeschätzte) untere Schranke von A das bisherige Minimum verbessert:
        A: Mask = S.pop();
        if lb < lb_estimate:
            # Bound, wenn sich A nicht weiter aufteilen lässt od. man A wie eine einelementige Option behandeln kann:
            if not A.splittable() or can_add_all or can_add_none:
                lb_estimate = lb;
                if can_add_all:
                    vector = A.pad_ones();
                elif can_add_none:
                    vector = A.pad_zeros();
                else:
                    vector = A;
            # Branch sonst
            else:
                B, C = A.split();
                S.push(B);
                # Nur dann C auf Stack legen, wenn mind. eine Möglichkeit in C die Kapazitätsschranke erfüllt:
                if sum(costs[C.indexes_one]) <= max_cost:
                    S.push(C);

    # Aspekte der Lösung speichern
    rucksack = vector.indexes_one; # Indexes von Items im Rucksack
    soln = Solution(
        vector = vector.decision,
        items = items[rucksack].tolist(),
        values = values[rucksack].tolist(),
        costs = costs[rucksack].tolist(),
    );

    # verbose output hier behandeln (irrelevant für Algorithmus):
    if verbose:
        repr = display_branch_and_bound(values=values, steps=logged_steps, order=order);
        repr_rucksack = display_rucksack(items=items[rucksack], costs=costs[rucksack], values=values[rucksack]);
        print('\x1b[1mLösung\x1b[0m');
        print('');
        print(repr);
        print('');
        print(f'Mask: {soln.vector}');
        print('Rucksack:');
        print(repr_rucksack);
        print('');

    # Lösung ausgeben
    return soln;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# AUXILIARY METHOD resort
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def get_sort_order(costs: np.ndarray, values: np.ndarray) -> List[int]:
    '''
    Sortiert Daten absteigend nach values/costs.
    '''
    n = len(costs);
    indexes = list(range(n));
    order = sorted(indexes, key=lambda i: -values[i]/costs[i]);
    return order;

def estimate_lower_bound(
    mask: Mask,
    max_cost: float,
    costs: np.ndarray,
    values: np.ndarray,
    items: np.ndarray,
) -> Tuple[float, List[float], bool]:
    '''
    Wenn partielle Information über den Rucksack festgelegt ist,
    kann man bei dem unbekannten Teil das Rucksack-Problem
    mit Greedy-Algorithmus »lösen«,
    um schnell eine gute Einschätzung zu bestimmen.

    NOTE: Diese Funktion wird `g(vector)` im Skript bezeichnet.
    '''
    indexes_one = mask.indexes_one;
    indexes_unset = mask.indexes_unset;
    n = len(mask);
    vector = np.zeros(shape=(n,), dtype=float);

    # Berechnungen bei Items mit bekanntem Status in Rucksack:
    value_rucksack = sum(values[indexes_one]);
    cost_rucksack = sum(costs[indexes_one]);
    vector[indexes_one] = 1;

    # Für Rest des Rucksacks (Items mit unbekanntem Status):
    cost_rest = max_cost - cost_rucksack;
    can_add_all = False;
    can_add_none = False;
    # Prüfe, ob man als Lösung alles/nichts hinzufügen kann:
    if len(indexes_unset) > 0 and sum(costs[indexes_unset]) <= cost_rest:
        can_add_all = True;
        vector[indexes_unset] = 1;
        value_rest = sum(values[indexes_unset]);
    elif len(indexes_unset) > 0 and min(costs[indexes_unset]) > cost_rest:
        can_add_none = True;
        vector[indexes_unset] = 0;
        value_rest = 0;
    # Sonst mit Greedy-Algorithmus lösen:
    # NOTE: Lösung ist eine Überschätzung des max-Wertes.
    else:
        soln_rest = rucksack_greedy_algorithm(
            max_cost = cost_rest, # <- Kapazität = Restgewicht
            costs = costs[indexes_unset],
            values = values[indexes_unset],
            items = items[indexes_unset],
            fractional = True,
            verbose = False,
        );
        value_rest = soln_rest.total_value;
        vector[indexes_unset] = soln_rest.vector;

    # Einschätzung des max-Wertes:
    value_max_est = value_rucksack + value_rest;

    # Ausgabe mit -1 multiplizieren (weil maximiert wird):
    return -value_max_est, vector.tolist(), can_add_all, can_add_none;
